"""Mapping of all channels available in PSS/E
"""


channel_map = {
    "ANGLE, machine relative rotor angle (degrees).": 1,
    "PELEC, machine electrical power (pu on SBASE).": 2,
    "QELEC, machine reactive power.": 3,
    "ETERM, machine terminal voltage (pu).": 4,
    "EFD, generator main field voltage (pu).": 5,
    "PMECH, turbine mechanical power (pu on MBASE).": 6,
    "SPEED, machine speed deviation from nominal (pu).": 7,
    "XADIFD, machine field current (pu).": 8,
    "ECOMP, voltage regulator compensated voltage (pu).": 9,
    "VOTHSG, stabilizer output signal (pu).": 10,
    "VREF, voltage regulator voltage setpoint (pu).": 11,
    "BSFREQ, bus pu frequency deviations.": 12,
    "VOLT, bus pu voltages (complex).": 13,
    "voltage and angle.": 14,
    "flow (P).": 15,
    "flow (P and Q).": 16,
    "flow (MVA).": 17,
    "apparent impedance (R and X).": 18,
    "ITERM.": 21,
    "machine apparent impedance.": 22,
    "VUEL, minimum excitation limiter output signal (pu).": 23,
    "VOEL, maximum excitation limiter output signal (pu).": 24,
    "PLOAD.": 25,
    "QLOAD.": 26,
    "GREF, turbine governor reference.": 27,
    "LCREF, turbine load control reference.": 28,
    "WVLCTY, wind velocity (m/s).": 29,
    "WTRBSP, wind turbine rotor speed deviation (pu).": 30,
    "WPITCH, pitch angle (degrees).": 31,
    "WAEROT, aerodynamic torque (pu on MBASE).": 32,
    "WROTRV, rotor voltage (pu on MBASE).": 33,
    "WROTRI, rotor current (pu on MBASE).": 34,
    "WPCMND, active power command from wind control (pu on MBASE).": 35,
    "WQCMND, reactive power command from wind control (pu on MBASE).": 36,
}
